// WL_AGENT.C

#include "wl_def.h"
#pragma hdrstop

/*
=============================================================================

                                LOCAL CONSTANTS

=============================================================================
*/

#define MAXMOUSETURN    10


#define MOVESCALE       150l
#define BACKMOVESCALE   100l
#define ANGLESCALE      20

/*
=============================================================================

                                GLOBAL VARIABLES

=============================================================================
*/



//
// player state info
//
int32_t         thrustspeed;

word            plux,pluy;          // player coordinates scaled to unsigned

short           anglefrac;

objtype        *LastAttacker;

/*
=============================================================================

                                                 LOCAL VARIABLES

=============================================================================
*/


void    T_Player (objtype *ob);
void    T_Attack (objtype *ob);
void    WarpPlayer (int facedir);

statetype   s_player = {false,0,0,(statefunc) T_Player,NULL,NULL};
statetype   s_attack = {false,0,0,(statefunc) T_Attack,NULL,NULL};

struct atkinf
{
    int8_t    tics,attack,frame;              // attack is 1 for gun, 2 for knife
} attackinfo[4][14] =
{
    { {6,0,1},{6,2,2},{6,0,3},{6,-1,4} },
    { {6,0,1},{6,1,2},{6,0,3},{6,-1,4} },
    { {3,0,1},{12,0,2},{12,1,3},{12,0,4},{12,0,5},{3,-1,6} },
    { {6,0,1},{6,1,2},{6,4,3},{6,-1,4} },
};

//===========================================================================

//----------

void Attack (void);
void Use (void);
void Search (objtype *ob);
void SelectWeapon (void);
void SelectItem (void);

//----------

boolean TryMove (objtype *ob);
void T_Player (objtype *ob);

void ClipMove (objtype *ob, int32_t xmove, int32_t ymove);

/*
=============================================================================

                                CONTROL STUFF

=============================================================================
*/

#ifdef USE_BESTWEAPON
// Checks to make sure you have ammo for weapon
// returns "true" if yes and "false" if no
int ChkAtkAmmo (int weapon)
{
   switch (weapon)
   {
      case wp_knife:
		  return 1;
      case wp_pistol:
           if (gamestate.ammo > 0)
			   return 1;
         break;
	  case wp_machinegun:
		  if (gamestate.shells > 0)
			  return 1;
		  break;
	  case wp_chaingun:
		  if (gamestate.ammo2 > 0)
			  return 1;
		  break;
   }

   return 0;
}
#endif

/*
======================
=
= CheckWeaponChange
=
= Keys 1-4 change weapons
=
======================
*/

void CheckWeaponChange (void)
{
	// Changed by Havoc
    int i, kbuttons[11]={sc_1,sc_2,sc_3,sc_4,sc_5,sc_6,sc_7,sc_8,sc_9,sc_0};

    for (i=0; i<NUMWEAPONS; i++)
    {
		if ((Keyboard[kbuttons[i]]) && (i != gamestate.weapon))
		{
			if (gamestate.bestweapon & (1<<i) && ChkAtkAmmo(i))
			{
				gamestate.changingweapon = true;
				gamestate.goingdown = true;
				gamestate.nextweapon = gamestate.chosenweapon = i;
				
				if(gamestate.chosenweapon != 0)
					SD_PlaySound(KNIGHTDEATHSND);

				DrawWeapon();
				DrawAmmo();

				IN_ClearKey(kbuttons[i]);
				
				return;
			}
		}
	}
}


/*
=======================
=
= ControlMovement
=
= Takes controlx,controly, and buttonstate[bt_strafe]
=
= Changes the player's angle and position
=
= There is an angle hack because when going 70 fps, the roundoff becomes
= significant
=
=======================
*/

void ControlMovement (objtype *ob)
{
    int32_t oldx,oldy;
    int     angle;
    int     angleunits;

    thrustspeed = 0;

    oldx = player->x;
    oldy = player->y;

    if(buttonstate[bt_strafeleft])
    {
        angle = ob->angle + ANGLES/4;
        if(angle >= ANGLES)
            angle -= ANGLES;
        if(buttonstate[bt_run])
            Thrust(angle, RUNMOVE * MOVESCALE * tics);
        else
            Thrust(angle, BASEMOVE * MOVESCALE * tics);
    }

    if(buttonstate[bt_straferight])
    {
        angle = ob->angle - ANGLES/4;
        if(angle < 0)
            angle += ANGLES;
        if(buttonstate[bt_run])
            Thrust(angle, RUNMOVE * MOVESCALE * tics );
        else
            Thrust(angle, BASEMOVE * MOVESCALE * tics);
    }

    //
    // side to side move
    //
    if (buttonstate[bt_strafe])
    {
        //
        // strafing
        //
        //
        if (controlx > 0)
        {
            angle = ob->angle - ANGLES/4;
            if (angle < 0)
                angle += ANGLES;
            Thrust (angle,controlx*MOVESCALE);      // move to left
        }
        else if (controlx < 0)
        {
            angle = ob->angle + ANGLES/4;
            if (angle >= ANGLES)
                angle -= ANGLES;
            Thrust (angle,-controlx*MOVESCALE);     // move to right
        }
    }
    else
    {
        //
        // not strafing
        //
        anglefrac += controlx;
        angleunits = anglefrac/ANGLESCALE;
        anglefrac -= angleunits*ANGLESCALE;
        ob->angle -= angleunits;

        if (ob->angle >= ANGLES)
            ob->angle -= ANGLES;
        if (ob->angle < 0)
            ob->angle += ANGLES;

    }

    //
    // forward/backwards move
    //
    if (controly < 0)
    {
        Thrust (ob->angle,-controly*MOVESCALE); // move forwards
    }
    else if (controly > 0)
    {
        angle = ob->angle + ANGLES/2;
        if (angle >= ANGLES)
            angle -= ANGLES;
        Thrust (angle,controly*BACKMOVESCALE);          // move backwards
    }

    if (gamestate.victoryflag)              // watching the BJ actor
        return;
}

/*
=============================================================================

                            STATUS WINDOW STUFF

=============================================================================
*/


/*
==================
=
= StatusDrawPic
=
==================
*/

void StatusDrawPic (unsigned x, unsigned y, unsigned picnum)
{
    LatchDrawPicScaledCoord ((screenWidth-scaleFactor*320)/16 + scaleFactor*x,
        screenHeight-scaleFactor*(STATUSLINES-y),picnum);
}

void StatusDrawFace(unsigned picnum)
{
    StatusDrawPic(17, 4, picnum);
    StatusDrawLCD(picnum);
}


/*
==================
=
= DrawFace
=
==================
*/

void DrawFace (void)
{
//    if(viewsize == 21 && ingame) return;
    if (SD_SoundPlaying() == GETGATLINGSND)
        StatusDrawFace(GOTGATLINGPIC);
    else if (gamestate.health)
    {
#ifdef SPEAR
        if (godmode)
            StatusDrawFace(GODMODEFACE1PIC+gamestate.faceframe);
        else if (gamestate.health <= 100)
#endif
            StatusDrawFace(FACE1APIC+3*((100-gamestate.health)/16)+gamestate.faceframe);
		else if (gamestate.health >= 101)
			StatusDrawFace(FACE1APIC+gamestate.faceframe);
    }
    else
    {
#ifndef SPEAR
        if (LastAttacker && LastAttacker->obclass == needleobj)
            StatusDrawFace(MUTANTBJPIC);
        else
#endif
            StatusDrawFace(FACE8APIC);
    }
}

/*
===============
=
= UpdateFace
=
= Calls draw face if time to change
=
===============
*/

int facecount = 0;
int facetimes = 0;

void UpdateFace (void)
{
    // don't make demo depend on sound playback
    if(demoplayback || demorecord)
    {
        if(facetimes > 0)
        {
            facetimes--;
            return;
        }
    }
    else if(SD_SoundPlaying() == GETGATLINGSND)
        return;

    facecount += tics;
    if (facecount > US_RndT())
    {
        gamestate.faceframe = (US_RndT()>>6);
        if (gamestate.faceframe==3)
            gamestate.faceframe = 1;

        facecount = 0;
        DrawFace ();
    }
}



/*
===============
=
= LatchNumber
=
= right justifies and pads with blanks
=
===============
*/

static void LatchNumber (int x, int y, unsigned width, int32_t number)
{
    unsigned length,c;
    char    str[20];

    ltoa (number,str,10);

    length = (unsigned) strlen (str);

    while (length<width)
    {
        StatusDrawPic (x,y,N_BLANKPIC);
        x++;
        width--;
    }

    c = length <= width ? 0 : length-width;

    while (c<length)
    {
        StatusDrawPic (x,y,str[c]-'0'+ N_0PIC);
        x++;
        c++;
    }
}


/*
==================
=
= DrawTime
=
==================
*/

void DrawTime (void)
{
	int  sec,min;

//	if(viewsize == 21 && ingame) return;
	
	if (gamestate.godmode)
	{
		sec = (gamestate.GodModeCount-gamestate.TimeCount)/70;

		if(sec <= 0)
		{
			gamestate.GodModeCount = 0;

			if(godmode)
			{
				godmode = 0;
				gamestate.godmode = 0;
			}
		}
	}
	else if (gamestate.countdown)
	{
		sec = (gamestate.CountDownCount-gamestate.TimeCount)/70;

		if(sec <= 0)
		{
			gamestate.CountDownCount = 0;

			if(countdown)
			{
				countdown = 0;
				gamestate.countdown = false;
			}
		}
	}
	else
	{
		// Time = Timecount / 70
		sec = gamestate.TimeCount/70;

		// 99 minutes max
		if (sec > 99*60)
			sec = 99*60;
	}
	min=sec/60;
	sec%=60;
	LatchNumber (11,22,1,(min/10));
	LatchNumber (12,22,1,(min%10));
	LatchNumber (14,22,1,(sec/10));
	LatchNumber (15,22,1,(sec%10));
}


/*
===============
=
= DrawHealth
=
===============
*/

void DrawHealth (void)
{
//    if(viewsize == 21 && ingame) return;
    LatchNumber (23,5,3,gamestate.health);
}


/*
===============
=
= TakeDamage
=
===============
*/

void TakeDamage (int points,objtype *attacker)
{
    LastAttacker = attacker;

    if (gamestate.victoryflag)
        return;
    if (gamestate.difficulty==gd_baby)
        points>>=2;

    if (!godmode)
        gamestate.health -= points;

    if (gamestate.health<=0)
    {
        gamestate.health = 0;
        playstate = ex_died;
        killerobj = attacker;
    }

    if (godmode != 2)
        StartDamageFlash (points);

    DrawHealth ();
    DrawFace ();

    //
    // MAKE BJ'S EYES BUG IF MAJOR DAMAGE!
    //
#ifdef SPEAR
    if (points > 30 && gamestate.health!=0 && !godmode/* && viewsize != 21*/)
    {
        StatusDrawFace(BJOUCHPIC);
        facecount = 0;
    }
#endif
}

/*
===============
=
= HealSelf
=
===============
*/

void HealSelf (int points)
{
    gamestate.health += points;

	switch(gamestate.healthexp)
	{
	case 1:
		{
			if (gamestate.health>100)
				gamestate.health = 100;
		}
		break;
	case 2:
		{
			if (gamestate.health>110)
				gamestate.health = 110;
		}
		break;
	case 3:
		{
			if (gamestate.health>120)
				gamestate.health = 120;
		}
		break;
	case 4:
		{
			if (gamestate.health>125)
				gamestate.health = 125;
		}
		break;
	}

    DrawHealth ();
    DrawFace ();
}


//===========================================================================


/*
===============
=
= DrawLevel
=
===============
*/

void DrawLevel (void)
{
//    if(viewsize == 21 && ingame) return;
    LatchNumber (1,22,2,gamestate.mapon+1);
}

//===========================================================================


/*
===============
=
= DrawEXP
=
===============
*/

void	DrawEXP (void)
{
//	if(viewsize == 21 && ingame) return;
	LatchNumber (4,22,3,gamestate.exp);
}


/*
===============
=
= GiveEXP
=
===============
*/

void	GiveEXP (int points)
{
	gamestate.exp += points;
	if (gamestate.exp>60)
		gamestate.exp = 60;

	DrawEXP ();
	SD_PlaySound (BONUS1UPSND);
}

//===========================================================================


/*
===============
=
= DrawLives
=
===============
*/

void DrawLives (void)
{
//    if(viewsize == 21 && ingame) return;
    LatchNumber (23,22,3,gamestate.lives);
}


/*
===============
=
= GiveExtraMan
=
===============
*/

void GiveExtraMan (void)
{
    if (gamestate.lives<10)
        gamestate.lives++;
    DrawLives ();
    SD_PlaySound (BONUS1UPSND);
}

//===========================================================================

/*
===============
=
= DrawScore
=
===============
*/

void DrawScore (void)
{
//    if(viewsize == 21 && ingame) return;
    LatchNumber (10,5,6,gamestate.score);
}

/*
===============
=
= GivePoints
=
===============
*/

void GivePoints (int32_t points)
{
    gamestate.score += points;
    while (gamestate.score >= gamestate.nextextra)
    {
        gamestate.nextextra += EXTRAPOINTS;
        GiveExtraMan ();
    }
    DrawScore ();
}

//===========================================================================

/*
==================
=
= DrawWeapon
=
==================
*/

void DrawWeapon (void)
{
//    if(viewsize == 21 && ingame) return;
    StatusDrawPic (34,23,KNIFEPIC+gamestate.weapon);
}


/*
==================
=
= DrawKeys
=
==================
*/

void DrawKeys (void)
{
//    if(viewsize == 21 && ingame) return;
    if (gamestate.keys & 1)
        StatusDrawPic (35,6,BLUEKEYPIC);
    else
        StatusDrawPic (35,6,NOKEYPIC);

    if (gamestate.keys & 2)
        StatusDrawPic (36,6,YELLOWKEYPIC);
    else
        StatusDrawPic (36,6,NOKEYPIC);

    if (gamestate.keys & 4)
        StatusDrawPic (37,6,BOSSPIC3PIC);
    else
        StatusDrawPic (37,6,NOKEYPIC);

    if (gamestate.keys & 8)
        StatusDrawPic (38,6,BOSSPIC4PIC);
    else
        StatusDrawPic (38,6,NOKEYPIC);
}

/*
==================
=
= GiveWeapon
=
==================
*/

void GiveWeapon (int weapon)
{
	// Changed by Havoc
#ifndef USE_BESTWEAPON
	GiveAmmo (6,wp_pistol);

    if (gamestate.bestweapon<weapon)
        gamestate.bestweapon = gamestate.weapon
        = gamestate.chosenweapon = weapon;
#else
	switch (weapon)
	{
	case wp_pistol:
		GiveAmmo (5,wp_pistol);
		break;
	case wp_machinegun:
		GiveAmmo (4,wp_machinegun);
		break;
	case wp_chaingun:
		GiveAmmo (12,wp_chaingun);
		break;
	}

	if ((gamestate.bestweapon & (1<<weapon)) == 0)
	{
		gamestate.changingweapon = true;
		gamestate.goingdown = true;
		gamestate.bestweapon |= (1<<weapon);
		gamestate.nextweapon = gamestate.chosenweapon = weapon;
	}
#endif

    DrawWeapon ();
	// Added by Havoc
#ifdef USE_BESTWEAPON
	DrawAmmo ();
#endif
}

//===========================================================================

/*
===============
=
= DrawAmmo
=
===============
*/

void DrawAmmo (void)
{
//    if(viewsize == 21 && ingame) return;
	switch(gamestate.chosenweapon)
	{
	case wp_pistol:
		LatchNumber (29,5,3,gamestate.ammo);
		break;
	case wp_machinegun:
		LatchNumber (29,5,3,gamestate.shells);
		break;
	case wp_chaingun:
		LatchNumber (29,5,3,gamestate.ammo2);
		break;
	}
}

/*
===============
=
= GiveAmmo
=
===============
*/

void GiveAmmo (int ammo, int weapon)
{
    if (!gamestate.ammo && !gamestate.ammo2 && !gamestate.shells)    // knife was out
    {
        if (!gamestate.attackframe)
        {
			gamestate.changingweapon = true;
			gamestate.goingdown = true;
            gamestate.nextweapon = gamestate.chosenweapon = weapon;
            DrawWeapon ();
        }
    }

	switch (weapon)
	{
	case wp_pistol:
		{
			gamestate.ammo += ammo;
			switch(gamestate.ammoexp)
			{
			case 1:
				{
					if (gamestate.ammo>100)
						gamestate.ammo = 100;
				}
				break;
			case 2:
				{
					if (gamestate.ammo>150)
						gamestate.ammo = 150;
				}
				break;
			case 3:
				{
					if (gamestate.ammo>200)
						gamestate.ammo = 200;
				}
				break;
			case 4:
				{
					if (gamestate.ammo>250)
						gamestate.ammo = 250;
				}
				break;
			}
		}
		break;
	case wp_machinegun:
		{
			gamestate.shells += ammo;
			switch(gamestate.ammoexp)
			{
			case 1:
				{
					if (gamestate.shells>100)
						gamestate.shells = 100;
				}
				break;
			case 2:
				{
					if (gamestate.shells>150)
						gamestate.shells = 150;
				}
				break;
			case 3:
				{
					if (gamestate.shells>200)
						gamestate.shells = 200;
				}
				break;
			case 4:
				{
					if (gamestate.shells>250)
						gamestate.shells = 250;
				}
				break;
			}
		}
		break;
	case wp_chaingun:
		{
			gamestate.ammo2 += ammo;
			switch(gamestate.ammoexp)
			{
			case 1:
				{
					if (gamestate.ammo2>100)
						gamestate.ammo2 = 100;
				}
				break;
			case 2:
				{
					if (gamestate.ammo2>150)
						gamestate.ammo2 = 150;
				}
				break;
			case 3:
				{
					if (gamestate.ammo2>200)
						gamestate.ammo2 = 200;
				}
				break;
			case 4:
				{
					if (gamestate.ammo2>250)
						gamestate.ammo2 = 250;
				}
				break;
			}
		}
		break;
	}
	DrawAmmo ();
}

//===========================================================================

/*
===============
=
= DrawMedkit
=
===============
*/

void	DrawMedkit (void)
{
//	if(viewsize == 21 && ingame) return;
	LatchNumber (29,22,3,gamestate.medkit);
}


/*
===============
=
= GiveMedkit
=
===============
*/

void	GiveMedkit (int medkit)
{
	gamestate.medkit += medkit;
	if (gamestate.medkit > 10)
		gamestate.medkit = 10;
	DrawMedkit ();
}

//===========================================================================

/*
==================
=
= GiveKey
=
==================
*/

void GiveKey (int key)
{
    gamestate.keys |= (1<<key);
    DrawKeys ();
}



/*
=============================================================================

                                MOVEMENT

=============================================================================
*/


/*
===================
=
= GetBonus
=
===================
*/
void GetBonus (statobj_t *check)
{
	switch (check->itemnumber)
	{
	case	bo_medkit:
		if (gamestate.medkit == 10)
			return;

		SD_PlaySound (HEALTH2SND);
		GiveMedkit (1);
		GetMessage("Picked up a portable first aid kit");
		break;

	case	bo_key1:
	case	bo_key2:
	case	bo_key3:
	case	bo_key4:
		GiveKey (check->itemnumber - bo_key1);
		switch(check->itemnumber)
		{
		case bo_key1:
			GetMessage("Picked up a blue keycard");
			break;
		case bo_key2:
			GetMessage("Picked up a yellow keycard");
			break;
		case bo_key3:
			GetMessage("Picked up a red keycard");
			break;
		case bo_key4:
			GetMessage("Picked up a green keycard");
			break;
		}
		SD_PlaySound (GETKEYSND);
		break;

	case	bo_children:
		SD_PlaySound (BONUS4SND);
		GivePoints (2000);
		gamestate.treasurecount++;
		GetMessage("Ugly children rescued");
		break;
	case	bo_cross:
		SD_PlaySound (BONUS1SND);
		GivePoints (1000);
		gamestate.treasurecount++;
		GetMessage("Picked up a money bag");
		break;
	case	bo_chalice:
		SD_PlaySound (BONUS2SND);
		GivePoints (100);
		gamestate.treasurecount++;
		GetMessage("Picked up a coin");
		break;
	case	bo_bible:
		SD_PlaySound (BONUS3SND);
		GivePoints (500);
		gamestate.treasurecount++;
		GetMessage("Picked up 5 coins");
		break;

	case	bo_clip:
		switch(gamestate.ammoexp)
		{
		case 1:
			{
				if (gamestate.ammo == 100)
					return;
			}
			break;
		case 2:
			{
				if (gamestate.ammo == 150)
					return;
			}
			break;
		case 3:
			{
				if (gamestate.ammo == 200)
					return;
			}
			break;
		case 4:
			{
				if (gamestate.ammo == 250)
					return;
			}
			break;
		}

		SD_PlaySound (GETAMMOSND);
		GiveAmmo (8,wp_pistol);
		GetMessage("Picked up a clip");
		break;

	case	bo_clip2:
		switch(gamestate.ammoexp)
		{
		case 1:
			{
				if (gamestate.shells == 100)
					return;
			}
			break;
		case 2:
			{
				if (gamestate.shells == 150)
					return;
			}
			break;
		case 3:
			{
				if (gamestate.shells == 200)
					return;
			}
			break;
		case 4:
			{
				if (gamestate.shells == 250)
					return;
			}
			break;
		}

		SD_PlaySound (GETAMMOSND);
		GiveAmmo (4,wp_machinegun);
		GetMessage("Picked up 4 shotgun shells");
		break;

#ifdef SPEAR
	case	bo_25clip:
		switch(gamestate.ammoexp)
		{
		case 1:
			{
				if (gamestate.ammo2 == 100)
					return;
			}
			break;
		case 2:
			{
				if (gamestate.ammo2 == 150)
					return;
			}
			break;
		case 3:
			{
				if (gamestate.ammo2 == 200)
					return;
			}
			break;
		case 4:
			{
				if (gamestate.ammo2 == 250)
					return;
			}
			break;
		}

		SD_PlaySound (GETAMMOBOXSND);
		GiveAmmo (25,wp_chaingun);
		GetMessage("Picked up a box of bullets");
		break;
#endif

	case	bo_machinegun:
		SD_PlaySound (GETMACHINESND);
		GiveWeapon (wp_machinegun);
		GetMessage("You got the shotgun!");
		break;
	case	bo_chaingun:
		SD_PlaySound (GETGATLINGSND);
		facetimes = 38;
		GiveWeapon (wp_chaingun);
		GetMessage("You got the MGX 2704!");
		
//		if(viewsize != 21)
		StatusDrawFace (GOTGATLINGPIC);
		
		facecount = 0;
		break;

	case	bo_fullheal:
		SD_PlaySound (BONUS1UPSND);
		HealSelf (99);
		GiveAmmo (25,wp_pistol);
		GiveAmmo (25,wp_machinegun);
		GiveAmmo (25,wp_chaingun);
		GiveExtraMan ();
		gamestate.treasurecount++;
		GetMessage("Picked up a backpack");
		break;

	case	bo_hostage:
		SD_PlaySound (BONUS4SND);
		GivePoints (3000);
		gamestate.treasurecount++;
		GetMessage("Mineworker rescued");
		break;

	case	bo_food:
		switch(gamestate.healthexp)
		{
		case 1:
			{
				if (gamestate.health == 100)
					return;
			}
			break;
		case 2:
			{
				if (gamestate.health == 110)
					return;
			}
			break;
		case 3:
			{
				if (gamestate.health == 120)
					return;
			}
			break;
		case 4:
			{
				if (gamestate.health == 125)
					return;
			}
			break;
		}

		SD_PlaySound (HEALTH1SND);
		HealSelf (10);
		GetMessage("Picked up a small first aid kit");
		break;

	case	bo_alpo:
		switch(gamestate.healthexp)
		{
		case 1:
			{
				if (gamestate.health == 100)
					return;
			}
			break;
		case 2:
			{
				if (gamestate.health == 110)
					return;
			}
			break;
		case 3:
			{
				if (gamestate.health == 120)
					return;
			}
			break;
		case 4:
			{
				if (gamestate.health == 125)
					return;
			}
			break;
		}

		SD_PlaySound (HEALTH1SND);
		HealSelf (4);
		GetMessage("Ate some food");
		break;

	case bo_godmode:
		SD_PlaySound (BONUS1UPSND);
		HealSelf (125);

		gamestate.GodModeCount = (15*70)+gamestate.TimeCount;
		godmode = 1;
		gamestate.godmode = 1;

		GetMessage("Picked up a syringe");
		break;

	case	bo_implant:
		if (gamestate.exp == 45)
			return;

		SD_PlaySound (BONUS1UPSND);
		GiveEXP (5);
		GetMessage("Picked up an implant");
		break;

	case	bo_diary1:
	case	bo_diary2:
	case	bo_diary3:
		// If this condition ever occur something must be very wrong!
		if(gamestate.diary == 3)
			return;

		SD_PlaySound (BONUS1UPSND);
		gamestate.diary++;
		GetMessage("Picked up a diary");
		break;

	case	bo_arm:
		if(gamestate.arm == 1)
			return;

		SD_PlaySound (BONUS1UPSND);
		gamestate.arm = true;
		GetMessage("Picked up a disembodied arm");
		break;

	case	bo_nuke:
		SD_PlaySound (PAINSCREAMSND);
		TakeDamage (25,NULL);
		break;

	case	bo_gibs:
		if (gamestate.health >= 10)
			return;

		HealSelf (1);
		GetMessage("Drank some water... tastes stale!");
		break;
	}

    StartBonusFlash ();
    check->shapenum = -1;                   // remove from list
}

/*
===================
=
= TryMove
=
= returns true if move ok
= debug: use pointers to optimize
===================
*/

boolean TryMove (objtype *ob)
{
    int         xl,yl,xh,yh,x,y;
    objtype    *check;
    int32_t     deltax,deltay;

    xl = (ob->x-PLAYERSIZE) >>TILESHIFT;
    yl = (ob->y-PLAYERSIZE) >>TILESHIFT;

    xh = (ob->x+PLAYERSIZE) >>TILESHIFT;
    yh = (ob->y+PLAYERSIZE) >>TILESHIFT;

#define PUSHWALLMINDIST PLAYERSIZE

    //
    // check for solid walls
    //
    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            check = actorat[x][y];
            if (check && !ISPOINTER(check))
            {
                if(tilemap[x][y]==64 && x==pwallx && y==pwally)   // back of moving pushwall?
                {
                    switch(pwalldir)
                    {
                        case di_north:
                            if(ob->y-PUSHWALLMINDIST<=(pwally<<TILESHIFT)+((63-pwallpos)<<10))
                                return false;
                            break;
                        case di_west:
                            if(ob->x-PUSHWALLMINDIST<=(pwallx<<TILESHIFT)+((63-pwallpos)<<10))
                                return false;
                            break;
                        case di_east:
                            if(ob->x+PUSHWALLMINDIST>=(pwallx<<TILESHIFT)+(pwallpos<<10))
                                return false;
                            break;
                        case di_south:
                            if(ob->y+PUSHWALLMINDIST>=(pwally<<TILESHIFT)+(pwallpos<<10))
                                return false;
                            break;
                    }
                }
                else return false;
            }
        }
    }

    //
    // check for actors
    //
    if (yl>0)
        yl--;
    if (yh<MAPSIZE-1)
        yh++;
    if (xl>0)
        xl--;
    if (xh<MAPSIZE-1)
        xh++;

    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            check = actorat[x][y];
            if (ISPOINTER(check) && check != player && (check->flags & FL_SHOOTABLE) && (check->obclass != turretobj) )
            {
                deltax = ob->x - check->x;
                if (deltax < -MINACTORDIST || deltax > MINACTORDIST)
                    continue;
                deltay = ob->y - check->y;
                if (deltay < -MINACTORDIST || deltay > MINACTORDIST)
                    continue;

                return false;
            }
        }
    }

    return true;
}


/*
===================
=
= ClipMove
=
===================
*/

void ClipMove (objtype *ob, int32_t xmove, int32_t ymove)
{
    int32_t    basex,basey;

    basex = ob->x;
    basey = ob->y;

    ob->x = basex+xmove;
    ob->y = basey+ymove;
    if (TryMove (ob))
        return;

#ifndef REMDEBUG
    if (noclip && ob->x > 2*TILEGLOBAL && ob->y > 2*TILEGLOBAL
        && ob->x < (((int32_t)(mapwidth-1))<<TILESHIFT)
        && ob->y < (((int32_t)(mapheight-1))<<TILESHIFT) )
        return;         // walk through walls
#endif

//    if (!SD_SoundPlaying())
//        SD_PlaySound (HITWALLSND);

    ob->x = basex+xmove;
    ob->y = basey;
    if (TryMove (ob))
        return;

    ob->x = basex;
    ob->y = basey+ymove;
    if (TryMove (ob))
        return;

    ob->x = basex;
    ob->y = basey;
}

//==========================================================================

/*
===================
=
= VictoryTile
=
===================
*/

void VictoryTile (void)
{
#ifndef SPEAR
    SpawnBJVictory ();
#endif

    gamestate.victoryflag = true;
}

/*
===================
=
= Thrust
=
===================
*/

// For player movement in demos exactly as in the original Wolf3D v1.4 source code
static fixed FixedByFracOrig(fixed a, fixed b)
{
    int sign = 0;
    if(b == 65536) b = 65535;
    else if(b == -65536) b = 65535, sign = 1;
    else if(b < 0) b = (-b), sign = 1;

    if(a < 0)
    {
        a = -a;
        sign = !sign;
    }
    fixed res = (fixed)(((int64_t) a * b) >> 16);
    if(sign)
        res = -res;
    return res;
}

void Thrust (int angle, int32_t speed)
{
    int32_t xmove,ymove;
    unsigned offset;

	if (switches.switch3)
	{
		if (gamestate.attackframe == 0)
		{
			if (gamestate.bobdir != 2)
				gamestate.bobber++;

			if (gamestate.bobdir == 2)
				gamestate.bobber--;

			if (gamestate.bobber == 30)
				gamestate.bobdir = 2;

			if (gamestate.bobber == 0)
				gamestate.bobdir = 1;
			if (gamestate.bobdir2 != 2)
				gamestate.bobber2++;
			
			if (gamestate.bobdir2 == 2)
				gamestate.bobber2--;

			if (gamestate.bobber2 == 30)
				gamestate.bobdir2 = 2;

			if (gamestate.bobber2 == -30)
				gamestate.bobdir2 = 1;
		}
	}

    //
    // ZERO FUNNY COUNTER IF MOVED!
    //
#ifdef SPEAR
    if (speed)
        funnyticount = 0;
#endif

    thrustspeed += speed;
    //
    // moving bounds speed
    //
    if (speed >= MINDIST*2)
        speed = MINDIST*2-1;

    xmove = DEMOCHOOSE_ORIG_SDL(
                FixedByFracOrig(speed, costable[angle]),
                FixedMul(speed,costable[angle]));
    ymove = DEMOCHOOSE_ORIG_SDL(
                -FixedByFracOrig(speed, sintable[angle]),
                -FixedMul(speed,sintable[angle]));

    ClipMove(player,xmove,ymove);

    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);

    offset = (player->tiley<<mapshift)+player->tilex;
    player->areanumber = *(mapsegs[0] + offset) -AREATILE;

    if (*(mapsegs[1] + offset) == EXITTILE)
	{
		happyend = true;
		playstate = ex_victorious;
	}
}


/*
=============================================================================

                                ACTIONS

=============================================================================
*/


/*
===============
=
= Cmd_Fire
=
===============
*/

void Cmd_Fire (void)
{
	if(gamestate.changingweapon == 1)
		return;

    buttonheld[bt_attack] = true;

    gamestate.weaponframe = 0;
	gamestate.bobber = gamestate.bobdir = gamestate.bobber2 = gamestate.bobdir2 = 0;

    player->state = &s_attack;

    gamestate.attackframe = 0;
    gamestate.attackcount =
        attackinfo[gamestate.weapon][gamestate.attackframe].tics;
    gamestate.weaponframe =
        attackinfo[gamestate.weapon][gamestate.attackframe].frame;
}

//===========================================================================

void PWCorrect (void)
{
	SD_PlaySound(DEATHSCREAM7SND);

	if(gamestate.mapon == 2 || gamestate.mapon == 4 || gamestate.mapon == 10)
		GetMessage("Turrets deactivated");
	else
		GetMessage("Security overridden");

	gamestate.securitylock = false;
}

void PWWrong (void)
{
	SD_PlaySound(DEATHSCREAM8SND);
	GetMessage("You need the correct password");
	gamestate.securitylock = true;
}

/*
===============
=
= Cmd_Use
=
===============
*/

void Cmd_Use (void)
{
    int     checkx,checky,doornum,dir;
    boolean elevatorok;
	statobj_t *statptr;    // Added by Havoc for interactive objects

    //
    // find which cardinal direction the player is facing
    //
    if (player->angle < ANGLES/8 || player->angle > 7*ANGLES/8)
    {
        checkx = player->tilex + 1;
        checky = player->tiley;
        dir = di_east;
        elevatorok = true;
    }
    else if (player->angle < 3*ANGLES/8)
    {
        checkx = player->tilex;
        checky = player->tiley-1;
        dir = di_north;
        elevatorok = false;
    }
    else if (player->angle < 5*ANGLES/8)
    {
        checkx = player->tilex - 1;
        checky = player->tiley;
        dir = di_west;
        elevatorok = true;
    }
    else
    {
        checkx = player->tilex;
        checky = player->tiley + 1;
        dir = di_south;
        elevatorok = false;
    }

	// Added by Havoc for interactive objects
	for (statptr = &statobjlist[0]; statptr !=laststatobj; statptr++)
	{ 
		if (statptr->tilex == checkx && statptr->tiley == checky && statptr->shapenum == SPR_STAT_1 && !buttonheld[bt_use])
		{
			switch (gamestate.healthexp)
			{
			case 1:
				{
					if (gamestate.health == 100)
						return;
				}
				break;
			case 2:
				{
					if (gamestate.health == 110)
						return;
				}
				break;
			case 3:
				{
					if (gamestate.health == 120)
						return;
				}
				break;
			case 4:
				{
					if (gamestate.health == 125)
						return;
				}
				break;
			}

			buttonheld[bt_use] = true;
			
			HealSelf(4);
			SD_PlaySound(HEALTH1SND);
			GetMessage("Ate some food");
			statptr->shapenum = SPR_STAT_2;
		}

		if (statptr->tilex == checkx && statptr->tiley == checky &&
		   (statptr->shapenum == SPR_STAT_64 || statptr->shapenum == SPR_STAT_77 || statptr->shapenum == SPR_STAT_78 || statptr->shapenum == SPR_STAT_79) &&
		   !buttonheld[bt_use])
		{
			buttonheld[bt_use] = true;

			ClearMemory ();

			VW_FadeOut();
			
			switch (statptr->shapenum)
			{
			case SPR_STAT_64:
				LogDiscScreens("1");
				break;

			case SPR_STAT_77:
				LogDiscScreens("2");
				break;

			case SPR_STAT_78:
				LogDiscScreens("3");
				break;

			case SPR_STAT_79:
				LogDiscScreens("4");
				break;
			}

			ClearMemory ();

			IN_ClearKeysDown();

			DrawPlayScreen ();
		}

		if (statptr->tilex == checkx && statptr->tiley == checky && statptr->shapenum == SPR_STAT_39 && !buttonheld[bt_use])
		{
			char password[6];
			boolean esc;

			buttonheld[bt_use] = true;
			
			fontnumber = 0;
			SETFONTCOLOR(0x0,15);

			if(gamestate.mapon == 7 && gamestate.electricity == 1)
			{
				CenterWindow(18, 2);
				US_PrintCentered("Error!");
				VW_UpdateScreen();
				IN_Ack();

				gamestate.electricity = false;
			}
			else if(gamestate.securitylock == 1)
			{
				CenterWindow (25,3);
				US_PrintCentered("Enter Password: ");

				VW_UpdateScreen();
				esc = !US_LineInput(px,py,password,NULL,true,5,0);

				if(!esc)
				{
					switch (gamestate.mapon)
					{
					case 1:
						if(strcmp(password,"94276")==0)
							PWCorrect();
						else
							PWWrong();
						break;
					case 2:
						if(strcmp(password,"12345")==0)
							PWCorrect();
						else
							PWWrong();
						break;
					case 4:
						if(strcmp(password,"29540")==0)
							PWCorrect();
						else
							PWWrong();
						break;
					case 7:
						PWWrong();
						break;
					case 10:
						if(strcmp(password,"68425")==0)
							PWCorrect();
						else
							PWWrong();
						break;
					case 12:
						if(strcmp(password,"24746")==0)
							PWCorrect();
						else
							PWWrong();
						break;
					}
				}
			}
		}

		if (statptr->tilex == checkx && statptr->tiley == checky && (statptr->shapenum == SPR_STAT_11 || statptr->shapenum == SPR_STAT_12) && !buttonheld[bt_use])
		{
			buttonheld[bt_use] = true;

			SD_PlaySound (DEATHSCREAM9SND);
		}

		if (statptr->tilex == checkx && statptr->tiley == checky && statptr->shapenum == SPR_STAT_81 && !buttonheld[bt_use])
		{
			buttonheld[bt_use] = true;
			
			SD_PlaySound (GETGATLINGSND);
			facetimes = 38;
			GiveWeapon (wp_chaingun);
			GetMessage("You got the MGX 2704!");
			
//			if(viewsize != 21)
			StatusDrawFace (GOTGATLINGPIC);

			facecount = 0;

			statptr->shapenum = SPR_STAT_85;
		}
	}

    doornum = tilemap[checkx][checky];
    if (*(mapsegs[1]+(checky<<mapshift)+checkx) == PUSHABLETILE)
    {
        //
        // pushable wall
        //

        PushWall (checkx,checky,dir);
        return;
    }
    if (!buttonheld[bt_use] && doornum == ELEVATORTILE && elevatorok)
    {
        //
        // use elevator
        //
        buttonheld[bt_use] = true;

        tilemap[checkx][checky]++;              // flip switch

		ThreeDRefresh();

		switch (*(mapsegs[1]+(player->tiley<<mapshift)+player->tilex))
		{
		case SECURITYTILE:
			if(gamestate.securitylock != false)
			{
				tilemap[checkx][checky]--;
				GetMessage("Security lock!");
			}
			else
				playstate = ex_completed;
			break;

		case WARPEASTTILE:
			tilemap[checkx][checky]--;
			WarpPlayer(EAST);
			break;

		case WARPWESTTILE:
			tilemap[checkx][checky]--;
			WarpPlayer(WEST);
			break;

		case WARPEASTTILE2:
			tilemap[checkx][checky]--;
			WarpPlayer(EAST2);
			break;

		case WARPWESTTILE2:
			tilemap[checkx][checky]--;
			WarpPlayer(WEST2);
			break;

		default:
			if(gamestate.mapon == 9 && gamestate.radar != 9)
			{
				tilemap[checkx][checky]--;
				GetMessage("Destroy all the radar equipment first!");
			}
			else if(gamestate.mapon == 13 && gamestate.diary != 3)
			{
				tilemap[checkx][checky]--;
				GetMessage("Find all the diaries first!");
			}
			else if (*(mapsegs[0]+(player->tiley<<mapshift)+player->tilex) == ALTELEVATORTILE)
				playstate = ex_secretlevel;
			else
				playstate = ex_completed;
			break;
		}
		
        SD_PlaySound (LEVELDONESND);
        SD_WaitSoundDone();
    }
	else if (!buttonheld[bt_use] && doornum == 14)
	{
	//
	// use upgrade unit
	//
		int upgrade=0;
		boolean esc;

		buttonheld[bt_use] = true;

		if (gamestate.exp >= 5)
		{
			fontnumber = 0;
			SETFONTCOLOR(0x0,15);

			CenterWindow (20,8);
			US_Print("UPGRADE UNIT\n1. Endurance (5 EXP)\n2. Strength (5 EXP)\n3. Accuracy (5 EXP)\n\nChoose your upgrade: ");

			VW_UpdateScreen();
			esc = !US_LineInput(px,py,str,NULL,true,1,0);

			if(!esc)
			{
				upgrade = atoi (str);

				switch(upgrade)
				{
				case 1:
					{
						if(gamestate.healthexp<4 && gamestate.exp >= 5)
						{
							gamestate.healthexp++;
							GiveEXP(-5);
						}
						else if(gamestate.exp<5)
						{
							GetMessage("Not enough implants for this upgrade");
						}
						else
							GetMessage("You reached the maximum upgrade level");
					}
					break;
				case 2:
					{
						if(gamestate.ammoexp<4 && gamestate.exp >= 5)
						{
							gamestate.ammoexp++;
							GiveEXP(-5);
						}
						else if(gamestate.exp<5)
						{
							GetMessage("Not enough implants for this upgrade");
						}
						else
							GetMessage("You reached the maximum upgrade level");
					}
					break;
				case 3:
					{
						if(gamestate.weaponexp<4 && gamestate.exp >= 5)
						{
							gamestate.weaponexp++;
							GiveEXP(-5);
						}
						else if(gamestate.exp<5)
						{
							GetMessage("Not enough implants for this upgrade");
						}
						else
							GetMessage("You reached the maximum upgrade level");
					}
					break;
				default:
					GetMessage("No upgrade");
				}
				DrawEXP();
			}
		}
		else
			GetMessage("You don't have enough implants");
	}
	else if (!buttonheld[bt_use] && doornum == 35)
	{
		buttonheld[bt_use] = true;

		switch(gamestate.ammoexp)
		{
		case 1:
			{
				if (gamestate.ammo == 100 && gamestate.ammo2 == 100 && gamestate.shells == 100)
					return;
			}
			break;
		case 2:
			{
				if (gamestate.ammo == 150 && gamestate.ammo2 == 150 && gamestate.shells == 150)
					return;
			}
			break;
		case 3:
			{
				if (gamestate.ammo == 200 && gamestate.ammo2 == 200 && gamestate.shells == 200)
					return;
			}
			break;
		case 4:
			{
				if (gamestate.ammo == 250 && gamestate.ammo2 == 250 && gamestate.shells == 250)
					return;
			}
			break;
		}
		
		tilemap[checkx][checky]++;
		SD_PlaySound(GETAMMOSND);
		GiveAmmo (25,wp_pistol);
		GiveAmmo (25,wp_chaingun);
		GiveAmmo (10,wp_machinegun);
		StartBonusFlash();
	}
	else if (!buttonheld[bt_use] && doornum == 37)
	{
		buttonheld[bt_use] = true;
		
		switch (gamestate.healthexp)
		{
		case 1:
			{
				if (gamestate.health == 100 && gamestate.medkit == 10)
					return;
			}
			break;
		case 2:
			{
				if (gamestate.health == 110 && gamestate.medkit == 10)
					return;
			}
			break;
		case 3:
			{
				if (gamestate.health == 120 && gamestate.medkit == 10)
					return;
			}
			break;
		case 4:
			{
				if (gamestate.health == 125 && gamestate.medkit == 10)
					return;
			}
			break;
		}
		
		tilemap[checkx][checky]--;
		SD_PlaySound(HEALTH2SND);
		HealSelf (25);
		if (gamestate.medkit < 10)
		{
			if (gamestate.medkit <= 8)
				GiveMedkit (2);
			else
				GiveMedkit (1);
		}
		StartBonusFlash();
	}
	else if (!buttonheld[bt_use] && (doornum == SWITCHTILE || doornum == SWITCHTILE+1))
	{
	//
	// use switch
	//
		buttonheld[bt_use] = true;
		
		if (doornum==SWITCHTILE)
		{
			tilemap[checkx][checky]++;			// flip switch

			if(gamestate.mapon != 12)
				gamestate.electricity=false;		// electricity is off

			if(gamestate.mapon == 4 || gamestate.mapon == 7 || gamestate.mapon == 10)
				GetMessage("Emergency power ON");
			else
				GetMessage("Security grid deactivated");
		}
		else
		{
			tilemap[checkx][checky]--;			// flip back

			if(gamestate.mapon != 12)
				gamestate.electricity=true;			// electricity is on

			if(gamestate.mapon == 4 || gamestate.mapon == 7 || gamestate.mapon == 10)
				GetMessage("Power line restored");
			else
				GetMessage("Security grid activated");
		}
		SD_PlaySound (LEVELDONESND);
		SD_WaitSoundDone();
	}
	else if (!buttonheld[bt_use] && MAPSPOT(checkx,checky,2) == 516)
	{
		buttonheld[bt_use] = true;

		if(gamestate.electricity == 1)
		{
			if(gamestate.arm == 1)
			{
				if(gamestate.mapon == 12)
					gamestate.electricity = false;
				else
					gamestate.securitylock = false;

				if(gamestate.mapon == 12)
					GetMessage("An error has occurred");
				else
					GetMessage("Access granted!");

				if(gamestate.mapon == 12)
					SD_PlaySound (DEATHSCREAM8SND);
				else
					SD_PlaySound (ACCESSGRANTEDSND);
			}
			else
			{
				GetMessage("Access denied!");
				SD_PlaySound (ACCESSDENIEDSND);
			}
		}
		else
		{
			if(gamestate.mapon == 7)
				gamestate.securitylock = true;

			GetMessage("Hmm, no power");
		}
	}
    else if (!buttonheld[bt_use] && doornum & 0x80)
    {
        buttonheld[bt_use] = true;
        OperateDoor (doornum & ~0x80);
    }
//    else
//        SD_PlaySound (DONOTHINGSND);
}

/*
=============================================================================

                                PLAYER CONTROL

=============================================================================
*/



/*
===============
=
= SpawnPlayer
=
===============
*/

void SpawnPlayer (int tilex, int tiley, int dir)
{
    player->obclass = playerobj;
    player->active = ac_yes;
    player->tilex = tilex;
    player->tiley = tiley;
    player->areanumber = (byte) *(mapsegs[0]+(player->tiley<<mapshift)+player->tilex);
    player->x = ((int32_t)tilex<<TILESHIFT)+TILEGLOBAL/2;
    player->y = ((int32_t)tiley<<TILESHIFT)+TILEGLOBAL/2;
    player->state = &s_player;
    player->angle = (1-dir)*90;
    if (player->angle<0)
        player->angle += ANGLES;
    player->flags = FL_NEVERMARK;
    Thrust (0,0);                           // set some variables

    InitAreas ();
}


//===========================================================================

/*
===============
=
= T_KnifeAttack
=
= Update player hands, and try to do damage when the proper frame is reached
=
===============
*/

void    KnifeAttack (objtype *ob)
{
    objtype *check,*closest;
    int32_t  dist;

    SD_PlaySound (ATKKNIFESND);
    // actually fire
    dist = 0x7fffffff;
    closest = NULL;
    for (check=ob->next; check; check=check->next)
    {
        if ( (check->flags & FL_SHOOTABLE) && (check->flags & FL_VISABLE)
            && abs(check->viewx-centerx) < shootdelta)
        {
            if (check->transx < dist)
            {
                dist = check->transx;
                closest = check;
            }
        }
    }

    if (!closest || dist > 0x18000l)
    {
        // missed
        return;
    }

    // hit something
    DamageActor (closest,US_RndT() >> 4);
}

void Medkit (void)
{
	if (gamestate.health >= 100)
		return;
	
	if (gamestate.health < 100)
	{
		SD_PlaySound (HEALTH2SND);
		HealSelf (10);

		GetMessage("Used a portable first aid kit");

		if(gamestate.health > 100)
			gamestate.health = 100;

		DrawHealth ();
		GiveMedkit (-1);

		IN_ClearKey(sc_X);
	}
}

void    GunAttack (objtype *ob)
{
    objtype *check,*closest,*oldclosest;
    int      damage;
    int      dx,dy,dist;
    int32_t  viewdist;

    switch (gamestate.weapon)
    {
        case wp_pistol:
            SD_PlaySound (ATKPISTOLSND);
            break;
        case wp_machinegun:
            SD_PlaySound (ATKMACHINEGUNSND);
            break;
        case wp_chaingun:
            SD_PlaySound (ATKGATLINGSND);
            break;
    }

    madenoise = true;

    //
    // find potential targets
    //
    viewdist = 0x7fffffffl;
    closest = NULL;

    while (1)
    {
        oldclosest = closest;

        for (check=ob->next ; check ; check=check->next)
        {
            if ((check->flags & FL_SHOOTABLE) && (check->flags & FL_VISABLE)
                && abs(check->viewx-centerx) < shootdelta)
            {
                if (check->transx < viewdist)
                {
                    viewdist = check->transx;
                    closest = check;
                }
            }
        }

        if (closest == oldclosest)
            return;                                         // no more targets, all missed

        //
        // trace a line from player to enemey
        //
        if (CheckLine(closest))
            break;
    }

    //
    // hit something
    //
    dx = ABS(closest->tilex - player->tilex);
    dy = ABS(closest->tiley - player->tiley);
    dist = dx>dy ? dx:dy;
    if (dist<2)
	{
		if(gamestate.weapon == wp_machinegun)
			damage = 35;
		else
			damage = US_RndT() / 4;
	}
    else if (dist<4)
        damage = US_RndT() / 6;
    else
    {
        if ( (US_RndT() / 12) < dist)           // missed
            return;
        damage = US_RndT() / 6;
    }

	switch (gamestate.weapon)
	{
	case wp_pistol:
		damage *= 0.50 * gamestate.weaponexp;
		break;
	case wp_machinegun:
		damage *= 0.75 * gamestate.weaponexp;
		break;
	case wp_chaingun:
		damage *= 0.25 * gamestate.weaponexp;
		break;
	}

    DamageActor (closest,damage);
}

//===========================================================================

/*
===============
=
= VictorySpin
=
===============
*/

void VictorySpin (void)
{
    int32_t    desty;

    if (player->angle > 270)
    {
        player->angle -= (short)(tics * 3);
        if (player->angle < 270)
            player->angle = 270;
    }
    else if (player->angle < 270)
    {
        player->angle += (short)(tics * 3);
        if (player->angle > 270)
            player->angle = 270;
    }

    desty = (((int32_t)player->tiley-5)<<TILESHIFT)-0x3000;

    if (player->y > desty)
    {
        player->y -= tics*4096;
        if (player->y < desty)
            player->y = desty;
    }
}


//===========================================================================

/*
===============
=
= T_Attack
=
===============
*/

void    T_Attack (objtype *ob)
{
    struct  atkinf  *cur;

    UpdateFace ();

    if (gamestate.victoryflag)              // watching the BJ actor
    {
        VictorySpin ();
        return;
    }

    if ( buttonstate[bt_use] && !buttonheld[bt_use] )
        buttonstate[bt_use] = false;

    if ( buttonstate[bt_attack] && !buttonheld[bt_attack])
        buttonstate[bt_attack] = false;

    ControlMovement (ob);
    if (gamestate.victoryflag)              // watching the BJ actor
        return;

    plux = (word) (player->x >> UNSIGNEDSHIFT);                     // scale to fit in unsigned
    pluy = (word) (player->y >> UNSIGNEDSHIFT);
    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);

    //
    // change frame and fire
    //
    gamestate.attackcount -= (short) tics;
    while (gamestate.attackcount <= 0)
    {
        cur = &attackinfo[gamestate.weapon][gamestate.attackframe];
        switch (cur->attack)
        {
#ifdef USE_BESTWEAPON
			// End of firing sequence - check ammo and fall back to best possible weapon with ammo 
            case -1:
                ob->state = &s_player;

                if (!ChkAtkAmmo (gamestate.weapon) && gamestate.chosenweapon != wp_knife)
                {
					for(int i=0; i<NUMWEAPONS; i++)
					{
						if(ChkAtkAmmo (i) && gamestate.bestweapon & (1<<i))
						{
							gamestate.changingweapon = true;
							gamestate.goingdown = true;
							gamestate.nextweapon = gamestate.chosenweapon = i;
						}
					}

                    DrawWeapon ();
					DrawAmmo ();
                }
                else
                {
                    if (gamestate.weapon != gamestate.chosenweapon)
                    {
						if(ChkAtkAmmo(gamestate.chosenweapon))
						{
							gamestate.changingweapon = true;
							gamestate.goingdown = true;
							gamestate.nextweapon = gamestate.chosenweapon;
						}

                        DrawWeapon ();
                    }
                }
                gamestate.attackframe = gamestate.weaponframe = 0;
                return;

            case 4:
                if (!ChkAtkAmmo(gamestate.weapon))
                    break;
                if (buttonstate[bt_attack])
                    gamestate.attackframe -= 2;
            case 1:
                if (!ChkAtkAmmo(wp_chaingun) && gamestate.weapon == wp_chaingun)
                {       // can only happen with chain gun
                    gamestate.attackframe++;
                    break;
                }

				if (ChkAtkAmmo(gamestate.weapon))
				{
					GunAttack (ob);

					if (!ammocheat)
					{
						switch (gamestate.weapon)
						{
						case wp_pistol:
							gamestate.ammo--;
							break;
						case wp_machinegun:
							gamestate.shells--;
							break;
						case wp_chaingun:
							gamestate.ammo2--;
							break;
						}
					}

					if (gamestate.weapon == wp_chaingun)
						Thrust (player->angle, -0x1000 - (gamestate.chosenweapon * 50));
					DrawAmmo ();
					break;
				}
				else
				{
					return;
				}

            case 2:
				KnifeAttack (ob);
                break;

            case 3:
                if (ChkAtkAmmo(gamestate.weapon) && buttonstate[bt_attack])
                    gamestate.attackframe -= 2;
                break;
#else
            case -1:
                ob->state = &s_player;
                if (!gamestate.ammo)
                {
                    gamestate.weapon = wp_knife;
                    DrawWeapon ();
                }
                else
                {
                    if (gamestate.weapon != gamestate.chosenweapon)
                    {
                        gamestate.weapon = gamestate.chosenweapon;
                        DrawWeapon ();
                    }
                }
                gamestate.attackframe = gamestate.weaponframe = 0;
                return;

            case 4:
                if (!gamestate.ammo)
                    break;
                if (buttonstate[bt_attack])
                    gamestate.attackframe -= 2;
            case 1:
                if (!gamestate.ammo)
                {       // can only happen with chain gun
                    gamestate.attackframe++;
                    break;
                }
                GunAttack (ob);
                if (!ammocheat)
                    gamestate.ammo--;
                DrawAmmo ();
                break;

            case 2:
                KnifeAttack (ob);
                break;

            case 3:
                if (gamestate.ammo && buttonstate[bt_attack])
                    gamestate.attackframe -= 2;
                break;
#endif
        }

        gamestate.attackcount += cur->tics;
        gamestate.attackframe++;
        gamestate.weaponframe =
            attackinfo[gamestate.weapon][gamestate.attackframe].frame;
    }
}



//===========================================================================

/*
===============
=
= T_Player
=
===============
*/

void    T_Player (objtype *ob)
{
    if (gamestate.victoryflag)              // watching the BJ actor
    {
        VictorySpin ();
        return;
    }

    UpdateFace ();
    CheckWeaponChange ();
	DrawTime ();

    if ( buttonstate[bt_use] )
        Cmd_Use ();

    if ( buttonstate[bt_attack] && !buttonheld[bt_attack])
        Cmd_Fire ();

    ControlMovement (ob);
    if (gamestate.victoryflag)              // watching the BJ actor
        return;

    plux = (word) (player->x >> UNSIGNEDSHIFT);                     // scale to fit in unsigned
    pluy = (word) (player->y >> UNSIGNEDSHIFT);
    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);
}

/*
===============
=
= WarpPlayer
=
===============
*/

void	WarpPlayer (int facedir)
{
	objtype	*check;
	
	switch(facedir)
	{
	case EAST:
		check = actorat[warpex][warpey];

		if(check > objlist && (check->flags & FL_SHOOTABLE))
			KillActor(check);
		
		SpawnPlayer(warpex,warpey,facedir);
		break;
	case WEST:
		check = actorat[warpwx][warpwy];
		
		if(check > objlist && (check->flags & FL_SHOOTABLE))
			KillActor(check);
		
		SpawnPlayer(warpwx,warpwy,facedir);
		break;
	case EAST2:
		check = actorat[warpex2][warpey2];

		if(check > objlist && (check->flags & FL_SHOOTABLE))
			KillActor(check);
		
		SpawnPlayer(warpex2,warpey2,facedir-1);
		break;
	case WEST2:
		check = actorat[warpwx2][warpwy2];
		
		if(check > objlist && (check->flags & FL_SHOOTABLE))
			KillActor(check);
		
		SpawnPlayer(warpwx2,warpwy2,facedir);
		break;
	}
	
	SD_PlaySound (BOSSACTIVESND);
	
	if(DigiMode != sds_Off)
		Delay (35); //adjust warp duration here
	else
		VW_WaitVBL(1*50);
	
	buttonstate[bt_attack] = false;
	buttonheld[bt_attack] = false;
	gamestate.attackframe = 0;
	gamestate.weaponframe = 0;

	ThreeDRefresh();
}